import json
import logging
import os

import dask
import dask.dataframe as dd

from ._statistics import (
    mean,
    std,
    max,
    histogram,
    run_dask_computation
)


logger = logging.getLogger(__name__)


def data_digest(filepath_in: str,
                filepath_out: str = "",
                compute_mean: bool = False,
                compute_std: bool = False,
                compute_max: bool = False,
                compute_histogram: int = -1
                ) -> dict:
    """Computes simple statistics for an input data set.

    Parameters
    ==========
    filepath_in : str
        File name + path of the .csv file.
    filepath_out : str, optional
        File name + path to store the output results.
        Default: "".
    compute_mean : bool, optional
        Flag to enable / disable the mean value statistic computation.
        Default: False.
    compute_std : bool, optional
        Flag to enable / disable the standard deviation statistic computation.
        Default: False.
    compute_max : bool, optional
        Flag to enable / disable the max value statistic computation.
        Default: False.
    compute_histogram : int, optional
        Flag to enable / disable the histogram computation. Positive values
        greater than 0 enable the computation of the histogram, using the
        passed value as the bucket size for the histogram. Otherwise,
        the histogram computation is disabled.
        Default: -1.

    Returns
    =======
    dict
        Computed statistics report.
    """
    data_ddf = load_csv_file(filepath_in)
    statistics: dict = set_statistics_computation(data_ddf,
                                                  compute_mean=compute_mean,
                                                  compute_std=compute_std,
                                                  compute_max=compute_max,
                                                  compute_histogram=compute_histogram)
    result: tuple = compute_statistics(statistics)
    report: dict = convert_tuple_to_dict(result)
    save_dict_to_json(report, filepath_out)
    return report


def load_csv_file(filepath: str,
                  timestamp_index: bool = True
                  ) -> dask.dataframe.core.DataFrame:
    """Loads a .csv file into a Dask DataFrame.

    Parameters
    ==========
    filepath : str
        File name + pathof the .csv file
    timestamp_index : bool, optional
        In case a ``timestamp`` column exists, use it as index for the dask dataframe.
        Default: True.

    Returns
    =======
    dask.dataframe.core.DataFrame
        Dask DataFrame.
    """
    assert os.path.exists(filepath), f"File name + path does not exist in the filesystem: {filepath}"
    filepath_ext: str = os.path.splitext(filepath)[1]
    assert filepath_ext == ".csv", f"Expecting a .csv file, got {filepath_ext}"

    ddf = dd.read_csv(filepath)

    if "timestamp" in ddf.columns:
        logger.debug("Timestamp column found! Setting it as a dataframe index.")
        ddf = ddf.set_index("timestamp")
    return ddf


def set_statistics_computation(ddf: dask.dataframe.core.DataFrame,
                               compute_mean: bool,
                               compute_std: bool,
                               compute_max: bool,
                               compute_histogram: int
                               ) -> dict:
    """Sets statistic computations over a dask dataframe.

    Parameters
    ==========
    ddf : dask.dataframe.core.DataFrame
        Dask DataFrame.
    compute_mean : bool
        Flag to enable / disable the mean value statistic computation.
    compute_std : bool
        Flag to enable / disable the standard deviation statistic computation.
    compute_max : bool
        Flag to enable / disable the maximum value statistic computation.
    compute_histogram : int
        Flag to enable / disable the histogram computation. Positive values
        greater than 0 enable the computation of the histogram, using the
        passed value as the bucket size for the histogram

    Returns
    =======
    dict
        Collection of statistics computations.
    """
    stats: dict = {}
    if compute_mean:
        stats["mean"] = mean(ddf)
    if compute_std:
        stats["stddev"] = std(ddf)
    if compute_max:
        stats["max"] = max(ddf)
    if compute_histogram > 0:
        stats["histogram"] = histogram(ddf, compute_histogram)
    return stats


def compute_statistics(stats: dict) -> tuple:
    """Runs the compute() step to gather the results of the dask DAG.

    Parameters
    ==========
    stats : dict
        Dictionary containing an aggregate of dask Series.

    Returns
    =======
    tuple
        Statistics results.
    """
    result: tuple = ()
    if any(stats):
        logger.debug("Run dask.compute().")
        result = run_dask_computation(stats)
    return result


def convert_tuple_to_dict(data: tuple) -> dict:
    """Converts the computed data as a tuple to a dictionary.

    Parameters
    ==========
    data : tuple
        Computed data.

    Returns
    =======
    dict
        Converted results.
    """
    output = {"columns": []}
    if any(data):
        converted = convert_computed_to_ordered_dict(data[0])
        for column in sorted(converted):
            output["columns"].append(converted[column])
    return output


def convert_computed_to_ordered_dict(d: dict) -> dict:
    """Converts a dictionary of pandas Series to a ordered dictionary.

    Parameters
    ==========
    d : dict
        Dictionary of panda Series.

    Returns
    =======
    dict
        Ordered dictionary.
    """
    out: dict = {}
    for key in d:
        s = dict(d[key])
        for column in s:
            if key == "histogram":
                hist = s[column]
                h = hist["h"]
                bins = hist["bins"]
                value = []
                for i in range(len(h)):
                    value.append({"from": bins[i].item(), "to": bins[i + 1].item(), "count": h[i].item()})
            else:
                value = s[column]
            try:
                out[column].update({key: value})
            except KeyError:
                out[column] = {"name": column, key: value}
    return out


def save_dict_to_json(data: dict, filepath: str) -> None:
    """Saves a python dictionary into a .json file.

    Parameters
    ==========
    data : dict
        Data to be stored to disk.
    filepath : str
        File name + path to store the data.
    """
    if any(filepath):
        filepath_ext: str = os.path.splitext(filepath)[1]
        assert filepath_ext == ".json", f"Expecting a .json file, got {filepath_ext}"
        logger.info(f"Saving results to file: {filepath}")
        with open(filepath, 'w') as outfile:
            json.dump(data, outfile)
