import dask
import dask.array as da
from dask.diagnostics import ProgressBar
import math
import numpy as np


def mean(ddf: dask.dataframe.core.DataFrame) -> dask.dataframe.core.Series:
    """Sets the mean value per column computation to a dask DAG.

    Parameters
    ==========
    ddf : dask.dataframe.core.DataFrame
        Dask DataFrame.

    Returns
    =======
    dask.dataframe.core.Series
        Dask Series of the aggregation operation.
    """
    return ddf.mean()


def std(ddf: dask.dataframe.core.DataFrame) -> dask.dataframe.core.Series:
    """Sets the standard deviation value per column computation to a dask DAG.

    Parameters
    ==========
    ddf : dask.dataframe.core.DataFrame
        Dask DataFrame.

    Returns
    =======
    dask.dataframe.core.Series
        Dask Series of the aggregation operation.
    """
    return ddf.std()


def max(ddf: dask.dataframe.core.DataFrame) -> dask.dataframe.core.Series:
    """Sets the max value per column computation to a dask DAG.

    Parameters
    ==========
    ddf : dask.dataframe.core.DataFrame
        Dask DataFrame.

    Returns
    =======
    dask.dataframe.core.Series
        Dask Series of the aggregation operation.
    """
    return ddf.max()


def histogram(ddf: dask.dataframe.core.DataFrame, bin_size: int) -> dict:
    """Computes a histogram for each column of a dask dataframe.

    Parameters
    ==========
    ddf : dask.dataframe.core.DataFrame
        Dask DataFrame.
    bin_size : int
        Size of a bin.

    Returns
    =======
    dict
        Dictionary of Dask Series of the aggregation operation.
    """
    range_values_per_column = compute_dataframe_range_per_column(ddf)

    output = {}
    for column in ddf.columns:
        if ddf[column].dtype not in [np.dtype("O"), np.dtype('<M8[ns]')]:
            range_min = parse_bin_size_multiple(range_values_per_column[column]["range_min"], bin_size)
            range_max = parse_bin_size_multiple(range_values_per_column[column]["range_max"], bin_size)
            bin_range = np.arange(min(0, range_min), range_max + 1e-8, bin_size)  # include range_max as well
            h, bins = da.histogram(convert_dask_dataframe_to_array(ddf[column]), bins=bin_range)
            output[column] = {"h": h, "bins": bins}
    return output


def compute_dataframe_range_per_column(ddf: dask.dataframe.core.DataFrame) -> dict:
    """Computes the range values for each numerical column of a dataframe a histogram from a dask dataframe.

    Parameters
    ==========
    ddf : dask.dataframe.core.DataFrame
        Dask DataFrame.

    Returns
    =======
    dict
        Dictionary of Dask Series of the aggregation operation.
    """
    range_values = {}
    for column in ddf.columns:
        if ddf[column].dtype not in [np.dtype("O"), np.dtype('<M8[ns]')]:
            range_values[column] = {
                "range_min": ddf[column].min(),
                "range_max": ddf[column].max()
            }
    result_range = run_dask_computation(range_values)
    return result_range[0]


def run_dask_computation(operation) -> tuple:
    """Computes a dask dag of operations."""
    with ProgressBar():
        result = dask.compute(operation)
    return result


def convert_dask_dataframe_to_array(ddf):
    """Converts a dask dataframe into a dask array."""
    return ddf.to_dask_array()


def parse_bin_size_multiple(value, size: int):
    """Parses a multiple of the size w.r.t. another value.

    Parameters
    ==========
    value : int | float
        Reference value.
    size : int
        Size of the bin.

    Returns
    =======
    int | float
        Multiple of the inut size bigger or smaller than value.
    """
    if value < 0:
        return math.floor(value / size) * size
    else:
        return math.ceil(value / size) * size
